/*
 Copyright (c) 2014,2015,2016 Michael Tross <michael@tross.org>

 This file is part of digitalStrom NetAtmo Virtual Device and needs
 libdSvDC from https://gitorious.digitalstrom.org/virtual-devices/libdsvdc.

 digitalStrom NetAtmo Virtual Device is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 libdsvdc is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pthread.h>

#include <libconfig.h>
#include <curl/curl.h>
#include <json.h>
#include <utlist.h>

#include <digitalSTROM/dsuid.h>
#include <dsvdc/dsvdc.h>

#include "netatmo.h"

#include "incbin.h"
INCBIN_EXTERN(IconStation16);
INCBIN_EXTERN(IconStation48);
INCBIN_EXTERN(IconIndoor16);
INCBIN_EXTERN(IconIndoor48);
INCBIN_EXTERN(IconRain16);
INCBIN_EXTERN(IconRain48);
INCBIN_EXTERN(IconWind16);
INCBIN_EXTERN(IconWind48);


void vdc_ping_cb(dsvdc_t *handle __attribute__((unused)), const char *dsuid, void *userdata __attribute__((unused)))
{
  int ret;
  vdc_report(LOG_NOTICE, "received ping for dsuid %s\n", dsuid);
  if (strcasecmp(dsuid, g_vdc_dsuid) == 0) {
    ret = dsvdc_send_pong(handle, dsuid);
    vdc_report(LOG_NOTICE, "sent pong for vdc %s / return code %d\n", dsuid, ret);
    return;
  }
  if (strcasecmp(dsuid, g_lib_dsuid) == 0) {
    ret = dsvdc_send_pong(handle, dsuid);
    vdc_report(LOG_NOTICE, "sent pong for lib-dsuid %s / return code %d\n", dsuid, ret);
    return;
  }
  netatmo_vdcd_t* dev;
  LL_FOREACH(devlist, dev)
  {
    if (strcasecmp(dsuid, dev->dsuidstring) == 0) {
      if (dev->present) {
        ret = dsvdc_send_pong(handle, dev->dsuidstring);
        vdc_report(LOG_NOTICE, "sent pong for device %s / return code %d\n", dsuid, ret);
      } else {
        vdc_report(LOG_NOTICE, "no pong for device %s, inactive / return code %d\n", dsuid, ret);
      }
      return;
    }
  }
  vdc_report(LOG_WARNING, "ping: no matching dsuid %s registered\n", dsuid);
}

void vdc_announce_device_cb(dsvdc_t *handle __attribute__((unused)), int code, void *arg, void *userdata __attribute__((unused)))
{
  vdc_report(LOG_INFO, "announcement of device %s returned code: %d\n", (char *) arg, code);
}

void vdc_announce_container_cb(dsvdc_t *handle __attribute__((unused)), int code, void *arg, void *userdata __attribute__((unused)))
{
  vdc_report(LOG_INFO, "announcement of container %s returned code: %d\n", (char *) arg, code);
}

void vdc_new_session_cb(dsvdc_t *handle, void *userdata)
{
    (void)userdata;
    int ret;
    ret = dsvdc_announce_container(handle,
                                   g_vdc_dsuid,
                                   (void *) g_vdc_dsuid,
                                   vdc_announce_container_cb);
    if (ret != DSVDC_OK)
    {
      vdc_report(LOG_WARNING, "dsvdc_announce_container returned error %d\n", ret);
      return;
    }

    vdc_report(LOG_INFO, "new session, container announced successfully\n");
}

void vdc_end_session_cb(dsvdc_t *handle, void *userdata)
{
    (void)userdata;
    int ret;

    vdc_report(LOG_WARNING, "end of session\n");
}

bool vdc_remove_cb(dsvdc_t *handle __attribute__((unused)), const char *dsuid, void *userdata)
{
  (void)userdata;
  vdc_report(LOG_INFO, "received remove for dsuid %s\n", dsuid);

  // TODO: what to do on remove?

  return true;
}

void vdc_blink_cb(dsvdc_t *handle __attribute__((unused)), char **dsuid, size_t n_dsuid,
    int32_t group, int32_t zone_id, void *userdata)
{
  (void) userdata;
  size_t n;

  for (n = 0; n < n_dsuid; n++) {
    vdc_report(LOG_NOTICE, "received blink for device %s: zone %d, group %d\n", *dsuid, zone_id, group);

    // TODO: what to do on blink?

  }
}

void vdc_setprop_cb(dsvdc_t *handle, const char *dsuid, dsvdc_property_t *property, const dsvdc_property_t *properties, void *userdata)
{
  (void) userdata;
  int ret;
  uint8_t code = DSVDC_ERR_NOT_IMPLEMENTED;
  size_t i;
  vdc_report(LOG_INFO, "set property request for dsuid \"%s\"\n", dsuid);

  /*
   * Properties for the VDC
   */
  if (strcasecmp(g_vdc_dsuid, dsuid) == 0) {
    for (i = 0; i < dsvdc_property_get_num_properties(properties); i++) {
      char *name;
      ret = dsvdc_property_get_name(properties, i, &name);
      if (ret != DSVDC_OK) {
        vdc_report(LOG_ERR, "setprop_cb: error getting property name\n");
        code = DSVDC_ERR_MISSING_DATA;
        break;
      }
      if (!name) {
        vdc_report(LOG_ERR, "setprop_cb: not handling wildcard properties\n");
        code = DSVDC_ERR_NOT_IMPLEMENTED;
        break;
      }

      vdc_report(LOG_INFO, "set request for name=\"%s\"\n", name);

      if (strcmp(name, "x-netatmo-username") == 0) {
        char* pstring;
        ret = dsvdc_property_get_string(properties, i, &pstring);
        if (ret != DSVDC_OK) {
          vdc_report(LOG_ERR, "setprop_cb: error getting property value from property %s\n", name);
          code = DSVDC_ERR_MISSING_DATA;
          break;
        }
        if (netatmo.username) {
          free(netatmo.username);
          netatmo.username = NULL;
        }
        if (pstring && strlen(pstring) > 0) {
          netatmo.username = strdup(pstring);
        }
        code = DSVDC_OK;
      }

      else if (strcmp(name, "x-netatmo-password") == 0) {
        char* pstring;
        ret = dsvdc_property_get_string(properties, i, &pstring);
        if (ret != DSVDC_OK) {
          vdc_report(LOG_ERR, "setprop_cb: error getting property value from property %s\n", name);
          code = DSVDC_ERR_MISSING_DATA;
          break;
        }
        if (netatmo.password) {
          free(netatmo.password);
          netatmo.password = NULL;
          if (g_access_token) {
            free(g_access_token);
            g_access_token = NULL;
          }
        }
        if (pstring && strlen(pstring) > 0) {
          netatmo.password = strdup(pstring);
        }
        code = DSVDC_OK;
      }

      else if (strcmp(name, "x-netatmo-refresh_token") == 0) {
        char* pstring;
        ret = dsvdc_property_get_string(properties, i, &pstring);
        if (ret != DSVDC_OK) {
          vdc_report(LOG_ERR, "setprop_cb: error getting property value from property %s\n", name);
          code = DSVDC_ERR_MISSING_DATA;
          break;
        }
        if (netatmo.refresh_token) {
          free(netatmo.refresh_token);
          netatmo.refresh_token = NULL;
        }
        if (pstring && strlen(pstring) > 0) {
          netatmo.refresh_token = strdup(pstring);
        }
        code = DSVDC_OK;
      }

      else if (strcmp(name, "x-netatmo-refresh_timeout") == 0) {
        char* pstring;
        ret = dsvdc_property_get_string(properties, i, &pstring);
        if (ret != DSVDC_OK) {
          vdc_report(LOG_ERR, "setprop_cb: error getting property value from property %s\n", name);
          code = DSVDC_ERR_MISSING_DATA;
          break;
        }
        if (pstring && strlen(pstring) > 0) {
          g_access_token_valid_until = time(NULL) + atoi(pstring) - 120;
        }
        code = DSVDC_OK;
      }

      else if (strcmp(name, "x-netatmo-access_token") == 0) {
        char* pstring;
        ret = dsvdc_property_get_string(properties, i, &pstring);
        if (ret != DSVDC_OK) {
          vdc_report(LOG_ERR, "setprop_cb: error getting property value from property %s\n", name);
          code = DSVDC_ERR_MISSING_DATA;
          break;
        }
        if (g_access_token) {
          free(g_access_token);
          g_access_token = NULL;
        }
        if (pstring && strlen(pstring) > 0) {
          g_access_token = strdup(pstring);
        }
        code = DSVDC_OK;
      }

      else if (strcmp(name, "x-netatmo-client_secret") == 0) {
        char* pstring;
        ret = dsvdc_property_get_string(properties, i, &pstring);
        if (ret != DSVDC_OK) {
          vdc_report(LOG_ERR, "setprop_cb: error getting property value from property %s\n", name);
          code = DSVDC_ERR_MISSING_DATA;
          break;
        }
        if (g_client_secret) {
          free(g_client_secret);
          g_client_secret = NULL;
        }
        if (pstring && strlen(pstring) > 0) {
          g_client_secret = strdup(pstring);
        }
        code = DSVDC_OK;
      }

      else if (strcmp(name, "x-netatmo-debug-level") == 0) {
        char* pstring;
        ret = dsvdc_property_get_string(properties, i, &pstring);
        if (ret != DSVDC_OK) {
          vdc_report(LOG_ERR, "setprop_cb: error getting property value from property %s\n", name);
          code = DSVDC_ERR_MISSING_DATA;
          break;
        }
        if (pstring && strlen(pstring) > 0) {
          vdc_set_debugLevel(atoi(pstring));
        }
        code = DSVDC_OK;
      }

      else {
        code = DSVDC_ERR_NOT_FOUND;
        break;
      }

      free(name);
    }

    if (code == DSVDC_OK) {
      write_config();
    }
  }

  dsvdc_send_set_property_response(handle, property, code);
}

void vdc_getprop_cb(dsvdc_t *handle, const char *dsuid, dsvdc_property_t *property, const dsvdc_property_t *query, void *userdata)
{
  (void) userdata;
  int ret;
  size_t i;
  char *name;
  netatmo_vdcd_t* dev;

  vdc_report(LOG_INFO, "get property for dsuid: %s\n", dsuid);

  /*
   * Properties for the VDC
   */
  if (strcasecmp(g_vdc_dsuid, dsuid) == 0) {
    for (i = 0; i < dsvdc_property_get_num_properties(query); i++) {

      int ret = dsvdc_property_get_name(query, i, &name);
      if (ret != DSVDC_OK) {
        vdc_report(LOG_ERR, "getprop_cb: error getting property name, abort\n");
        dsvdc_send_get_property_response(handle, property);
        return;
      }
      if (!name) {
        vdc_report(LOG_ERR, "getprop_cb: not yet handling wildcard properties\n");
        dsvdc_send_get_property_response(handle, property);
        return;
      }
      vdc_report(LOG_NOTICE, "get request name=\"%s\"\n", name);

      if (strcmp(name, "hardwareGuid") == 0) {
        char info[256];
        char buffer[32];
        size_t n;

        strcpy(info, "macaddress:");
        sprintf(buffer, "%02x:%02x:%02x:%02x:%02x:%02x", netatmo.base.bssid[5], netatmo.base.bssid[4],
            netatmo.base.bssid[3], netatmo.base.bssid[2], netatmo.base.bssid[1], netatmo.base.bssid[0]);
        strcat(info, buffer);
        dsvdc_property_add_string(property, name, info);

      } else if (strcmp(name, "modelUID") == 0) {
        dsvdc_property_add_string(property, name, "NetAtmo Weather Station");

      } else if (strcmp(name, "modelGuid") == 0) {
        dsvdc_property_add_string(property, name, "0.0.1.0");

      } else if (strcmp(name, "oemGuid") == 0) {
        dsvdc_property_add_string(property, name, "0.0.1.1");

      } else if (strcmp(name, "vendorId") == 0) {
        dsvdc_property_add_string(property, name, "michael@tross.org");

      } else if (strcmp(name, "name") == 0) {
        char info[256];
        strcpy(info, "NetAtmo Weather Station ");
        strcat(info, netatmo.base.station_name);
        dsvdc_property_add_string(property, name, info);

      } else if (strcmp(name, "model") == 0) {
        dsvdc_property_add_string(property, name, "NetAtmo Weather Station by mtr");

      } else if (strcmp(name, "capabilities") == 0) {
        dsvdc_property_t *reply;
        ret = dsvdc_property_new(&reply);
        if (ret != DSVDC_OK) {
          vdc_report(LOG_ERR, "failed to allocate reply property for %s\n", name);
          free(name);
          continue;
        }
        dsvdc_property_add_bool(reply, "metering", false);
        dsvdc_property_add_property(property, name, &reply);

      } else if (strcmp(name, "configURL") == 0) {
        dsvdc_property_add_string(property, name, "/add-ons/vdc-ui-netatmo/");

      /* user properties: user name, client_id, status */

      } else if (strcmp(name, "x-netatmo-username") == 0) {
        if (netatmo.username != NULL) {
          dsvdc_property_add_string(property, name, netatmo.username);
        } else {
          dsvdc_property_add_string(property, name, "");
        }

      } else if (strcmp(name, "x-netatmo-client-id") == 0) {
        dsvdc_property_add_string(property, name, g_client_id);

      } else if (strcmp(name, "x-netatmo-connection-status") == 0) {
        dsvdc_property_add_bool(property, name, g_access_token != NULL);

      }

      free(name);
    }

    dsvdc_send_get_property_response(handle, property);
    return;
  }

  pthread_mutex_lock(&g_network_mutex);
  LL_FOREACH(devlist, dev)
  {
    if (strcasecmp(dsuid, dev->dsuidstring) == 0) {
      break;
    }
  }
  pthread_mutex_unlock(&g_network_mutex);

  if (dev == NULL) {
    vdc_report(LOG_WARNING, "get property: unhandled dsuid %s\n", dsuid);
    dsvdc_property_free(property);
    return;
  }

  /*
   * Properties for the VDSD's
   */
  pthread_mutex_lock(&g_network_mutex);
  for (i = 0; i < dsvdc_property_get_num_properties(query); i++) {

    int ret = dsvdc_property_get_name(query, i, &name);
    if (ret != DSVDC_OK) {
      vdc_report(LOG_ERR, "getprop_cb: error getting property name, abort\n");
      dsvdc_send_get_property_response(handle, property);
      pthread_mutex_unlock(&g_network_mutex);
      return;
    }
    if (!name) {
      vdc_report(LOG_ERR, "getprop_cb: not yet handling wildcard properties\n");
      //dsvdc_send_property_response(handle, property);
      continue;
    }
    vdc_report(LOG_NOTICE, "get request name=\"%s\"\n", name);

    if (strcmp(name, "primaryGroup") == 0) {
      dsvdc_property_add_uint(property, "primaryGroup", 8);

    } else if (strcmp(name, "buttonInputDescriptions") == 0) {

    } else if (strcmp(name, "buttonInputSettings") == 0) {

    } else if (strcmp(name, "outputDescription") == 0) {

    } else if (strcmp(name, "outputSettings") == 0) {

    } else if (strcmp(name, "channelDescriptions") == 0) {

    } else if (strcmp(name, "channelSettings") == 0) {

    } else if (strcmp(name, "binaryInputDescriptions") == 0) {
      if ((strcmp(dev->mod->type, "NAModule1") == 0) ||
          (strcmp(dev->mod->type, "NAModule2") == 0) ||
          (strcmp(dev->mod->type, "NAModule3") == 0) ||
          (strcmp(dev->mod->type, "NAModule4") == 0)) {

        dsvdc_property_t *reply;
        ret = dsvdc_property_new(&reply);
        if (ret != DSVDC_OK) {
          vdc_report(LOG_ERR, "failed to allocate reply property for %s\n", name);
          free(name);
          continue;
        }

        dsvdc_property_t *nProp;
        if (dsvdc_property_new(&nProp) != DSVDC_OK) {
          break;
        }
        dsvdc_property_add_string(nProp, "name", "Battery Status");
        dsvdc_property_add_uint(nProp, "sensorFunction", 12);
        dsvdc_property_add_double(nProp, "updateInterval", 60 * 5);
        dsvdc_property_add_property(reply, "0", &nProp);

        dsvdc_property_add_property(property, name, &reply);
      }

    } else if (strcmp(name, "binaryInputSettings") == 0) {
      if ((strcmp(dev->mod->type, "NAModule1") == 0) ||
          (strcmp(dev->mod->type, "NAModule2") == 0) ||
          (strcmp(dev->mod->type, "NAModule3") == 0) ||
          (strcmp(dev->mod->type, "NAModule4") == 0)) {

        dsvdc_property_t *reply;
        ret = dsvdc_property_new(&reply);
        if (ret != DSVDC_OK) {
          vdc_report(LOG_ERR, "failed to allocate reply property for %s\n", name);
          free(name);
          continue;
        }

        dsvdc_property_t *nProp;
        if (dsvdc_property_new(&nProp) != DSVDC_OK) {
          break;
        }
        dsvdc_property_add_uint(nProp, "group", 8);
        dsvdc_property_add_uint(nProp, "sensorFunction", 12);
        dsvdc_property_add_property(reply, "0", &nProp);

        dsvdc_property_add_property(property, name, &reply);
      }

    } else if (strcmp(name, "sensorDescriptions") == 0) {
      dsvdc_property_t *reply;
      ret = dsvdc_property_new(&reply);
      if (ret != DSVDC_OK) {
        vdc_report(LOG_ERR, "failed to allocate reply property for %s\n", name);
        free(name);
        continue;
      }

      int n, sensorUsage, sensorType;
      char sensorName[64];
      char sensorIndex[64];

      if ((strcmp(dev->mod->type, "NAModule1") == 0) ||
          (strcmp(dev->mod->type, "NAModule2") == 0) ||
          (strcmp(dev->mod->type, "NAModule3") == 0)) {
        sensorUsage = 2; // Outdoor
      } else {
        sensorUsage = 1; // Indoor
      }

      if (strcmp(dev->mod->type, "NAModule2") == 0) {
        struct windTypeList { char* typeName; int typeId; };
        struct windTypeList wl[] = {
            { "WindAngle", 19 },
            { "WindStrength", 13 },
            { "GustAngle", 23 },
            { "GustStrength", 24 },
            { NULL, 0 }
        };
        for (n = 0; wl[n].typeName != NULL; n++) {
          snprintf(sensorIndex, 64, "%d", n);
          snprintf(sensorName, 64, "%s-%s", dev->mod->name, wl[n].typeName);

          dsvdc_property_t *nProp;
          if (dsvdc_property_new(&nProp) != DSVDC_OK) {
            vdc_report(LOG_ERR, "failed to allocate reply property for %s/%s\n", name, sensorName);
            break;
          }
          dsvdc_property_add_string(nProp, "name", sensorName);
          dsvdc_property_add_uint(nProp, "sensorType", wl[n].typeId);
          dsvdc_property_add_uint(nProp, "sensorUsage", sensorUsage);
          dsvdc_property_add_double(nProp, "aliveSignInterval", 300);
          dsvdc_property_add_property(reply, sensorIndex, &nProp);

          vdc_report(LOG_INFO, "sensorDescription: dsuid %s sensor %d: %s type %d usage %d\n", dsuid, n, sensorName,
              sensorType, sensorUsage);
        }

      } else if (strcmp(dev->mod->type, "NAModule3") == 0) {
        n = 0;
        snprintf(sensorIndex, 64, "%d", n);
        snprintf(sensorName, 64, "%s-%s", dev->mod->name, dev->mod->values[n].data_type);

        dsvdc_property_t *nProp;
        if (dsvdc_property_new(&nProp) != DSVDC_OK) {
          vdc_report(LOG_ERR, "failed to allocate reply property for %s/%s\n", name, sensorName);
          break;
        }
        dsvdc_property_add_string(nProp, "name", sensorName);
        dsvdc_property_add_uint(nProp, "sensorType", 21);
        dsvdc_property_add_uint(nProp, "sensorUsage", sensorUsage);
        dsvdc_property_add_double(nProp, "aliveSignInterval", 300);
        dsvdc_property_add_property(reply, sensorIndex, &nProp);

        vdc_report(LOG_INFO, "sensorDescription: dsuid %s sensor %d: %s type %d usage %d\n", dsuid, n, sensorName,
            sensorType, sensorUsage);

      } else {
        for (n = 0; n < dev->mod->values_num; n++) {
          if (strcmp(dev->mod->values[n].data_type, "Temperature") == 0) {
            sensorType = 1;
          } else if (strcmp(dev->mod->values[n].data_type, "Humidity") == 0) {
            sensorType = 2;
          } else if (strcmp(dev->mod->values[n].data_type, "CO2") == 0) {
            sensorType = 22;
          } else if (strcmp(dev->mod->values[n].data_type, "Pressure") == 0) {
            sensorType = 18;
          } else if (strcmp(dev->mod->values[n].data_type, "Noise") == 0) {
            sensorType = 20;
          } else {
            sensorType = 253;
          }
          snprintf(sensorName, 64, "%s-%s", dev->mod->name, dev->mod->values[n].data_type);

          dsvdc_property_t *nProp;
          if (dsvdc_property_new(&nProp) != DSVDC_OK) {
            vdc_report(LOG_ERR, "failed to allocate reply property for %s/%s\n", name, sensorName);
            break;
          }
          dsvdc_property_add_string(nProp, "name", sensorName);
          dsvdc_property_add_uint(nProp, "sensorType", sensorType);
          dsvdc_property_add_uint(nProp, "sensorUsage", sensorUsage);
          dsvdc_property_add_double(nProp, "aliveSignInterval", 300);

          snprintf(sensorIndex, 64, "%d", n);
          dsvdc_property_add_property(reply, sensorIndex, &nProp);

          vdc_report(LOG_INFO, "sensorDescription: dsuid %s sensor %d: %s type %d usage %d\n", dsuid, n, sensorName,
              sensorType, sensorUsage);
        }
      }

      dsvdc_property_add_property(property, name, &reply);

    } else if (strcmp(name, "sensorSettings") == 0) {
      dsvdc_property_t *reply;
      ret = dsvdc_property_new(&reply);
      if (ret != DSVDC_OK) {
        vdc_report(LOG_ERR, "failed to allocate reply property for %s\n", name);
        free(name);
        continue;
      }

      char sensorIndex[64];
      int n;
      for (n = 0; n < dev->mod->values_num; n++) {
        dsvdc_property_t *nProp;
        if (dsvdc_property_new(&nProp) != DSVDC_OK) {
          vdc_report(LOG_ERR, "failed to allocate reply property for %s\n", name);
          break;
        }
        dsvdc_property_add_uint(nProp, "group", 8);
        dsvdc_property_add_uint(nProp, "minPushInterval", 300);
        dsvdc_property_add_double(nProp, "changesOnlyInterval", 300);

        snprintf(sensorIndex, 64, "%d", n);
        dsvdc_property_add_property(reply, sensorIndex, &nProp);
      }
      dsvdc_property_add_property(property, name, &reply);

    } else if (strcmp(name, "sensorStates") == 0) {
      dsvdc_property_t *reply;
      ret = dsvdc_property_new(&reply);
      if (ret != DSVDC_OK) {
        vdc_report(LOG_ERR, "failed to allocate reply property for %s\n", name);
        free(name);
        continue;
      }

      int idx, n;
      char* sensorIndex;
      dsvdc_property_t *sensorRequest;
      dsvdc_property_get_property_by_index(query, 0, &sensorRequest);
      if (dsvdc_property_get_name(sensorRequest, 0, &sensorIndex) != DSVDC_OK) {
        vdc_report(LOG_WARNING, "sensorStates: could not parse index\n");
        idx = -1;
      } else {
        idx = strtol(sensorIndex, NULL, 10);
      }
      dsvdc_property_free(sensorRequest);

      time_t now = time(NULL);
      netatmo_get_values(dev->mod);

      for (n = 0; n < dev->mod->values_num; n++) {
        if (idx >= 0 && idx != n) {
          continue;
        }

        dsvdc_property_t *nProp;
        if (dsvdc_property_new(&nProp) != DSVDC_OK) {
          vdc_report(LOG_ERR, "failed to allocate reply property for %s\n", name);
          break;
        }

        double val = dev->mod->values[n].value;

        dsvdc_property_add_double(nProp, "value", val);
        dsvdc_property_add_int(nProp, "age", now - dev->mod->values[n].last_query);
        dsvdc_property_add_int(nProp, "error", 0);

        char replyIndex[32];
        snprintf(replyIndex, 64, "%d", n);
        dsvdc_property_add_property(reply, replyIndex, &nProp);
      }
      dsvdc_property_add_property(property, name, &reply);

    } else if (strcmp(name, "binaryInputStates") == 0) {
      dsvdc_property_t *reply;
      ret = dsvdc_property_new(&reply);
      if (ret != DSVDC_OK) {
        vdc_report(LOG_ERR, "failed to allocate reply property for %s\n", name);
        free(name);
        continue;
      }

      int idx, n;
      char* sensorIndex;
      dsvdc_property_t *sensorRequest;
      dsvdc_property_get_property_by_index(query, 0, &sensorRequest);
      if (dsvdc_property_get_name(sensorRequest, 0, &sensorIndex) != DSVDC_OK) {
        vdc_report(LOG_WARNING, "binaryInputStates: could not parse index\n");
        idx = -1;
      } else {
        idx = strtol(sensorIndex, NULL, 10);
      }
      dsvdc_property_free(sensorRequest);

      dsvdc_property_t *nProp;
      if (((idx < 0) || (idx == 0)) && (dsvdc_property_new(&nProp) == DSVDC_OK)) {

        bool val;
        if (dev->mod->battery_percent > 0) {
          val = dev->mod->battery_percent <= 25;
        } else if (strcmp(dev->mod->type, "NAModule1") == 0) {
          val = dev->mod->battery_vp <= 4000; // OUTDOOR
        } else if (strcmp(dev->mod->type, "NAModule2") == 0) {
          val = dev->mod->battery_vp <= 4360; // WIND
        } else if (strcmp(dev->mod->type, "NAModule3") == 0) {
          val = dev->mod->battery_vp <= 4000; // RAIN
        } else if (strcmp(dev->mod->type, "NAModule4") == 0) {
          val = dev->mod->battery_vp <= 4560; // INDOOR
        } else {
          val = 0;
        }

        dsvdc_property_add_bool(nProp, "value", val);
        dsvdc_property_add_int(nProp, "age", 0);
        dsvdc_property_add_int(nProp, "error", 0);

        dsvdc_property_add_property(reply, "0", &nProp);
      }
      dsvdc_property_add_property(property, name, &reply);

    } else if (strcmp(name, "name") == 0) {
      dsvdc_property_add_string(property, name, dev->mod->name);

    } else if (strcmp(name, "type") == 0) {
      dsvdc_property_add_string(property, name, "vDSD");

    } else if (strcmp(name, "model") == 0) {
      char info[256];
      if (strcmp(dev->mod->type, "NAModule1") == 0) {
        strcpy(info, "Outdoor");  // raingauge and outdoor module
      } else if (strcmp(dev->mod->type, "NAModule2") == 0) {
        strcpy(info, "Wind Gauge"); // wind modules
      } else if (strcmp(dev->mod->type, "NAModule3") == 0) {
        strcpy(info, "Rain Gauge"); // rain modules
      } else if (strcmp(dev->mod->type, "NAModule4") == 0) {
        strcpy(info, "Indoor"); // indoor modules
      } else if (strcmp(dev->mod->type, "NAMain") == 0) {
        strcpy(info, "Base Station"); // main modules
      } else {
        strcpy(info, "Unknown");
      }
      dsvdc_property_add_string(property, name, info);

    } else if (strcmp(name, "modelFeatures") == 0) {
      dsvdc_property_t *nProp;
      dsvdc_property_new(&nProp);
      dsvdc_property_add_bool(nProp, "outmode", false);
      dsvdc_property_add_bool(nProp, "otypeconfig", false);
      dsvdc_property_add_property(property, name, &nProp);

    } else if (strcmp(name, "modelUID") == 0) {
      dsvdc_property_add_string(property, name, dev->mod->type);

    } else if (strcmp(name, "vendorGuid") == 0) {
      char info[256];
      strcpy(info, "NetAtmo vDC ");
      strcat(info, dev->id);
      dsvdc_property_add_string(property, name, info);

    } else if (strcmp(name, "hardwareVersion") == 0) {
      dsvdc_property_add_string(property, name, "0.0.0");

    } else if (strcmp(name, "configURL") == 0) {
      dsvdc_property_add_string(property, name, "/add-ons/vdc-ui-netatmo/");

    } else if (strcmp(name, "hardwareModelGuid") == 0) {
      dsvdc_property_add_string(property, name, "gs1:3700730500111");

    } else if (strcmp(name, "hardwareGuid") == 0) {
      char info[256];
      char buffer[32];
      size_t n;

      strcpy(info, "macaddress:");
      sprintf(buffer, "%02x:%02x:%02x:%02x:%02x:%02x",
          dev->mod->bid[0], dev->mod->bid[1], dev->mod->bid[2],
          dev->mod->bid[3], dev->mod->bid[4], dev->mod->bid[5]);
      strcat(info, buffer);
      dsvdc_property_add_string(property, name, info);

    } else if (strcmp(name, "deviceIcon16") == 0) {
      if (strcmp(dev->mod->type, "NAModule1") == 0) {
        dsvdc_property_add_bytes(property, name, gIconIndoor16Data, gIconIndoor16Size);
      } else if (strcmp(dev->mod->type, "NAModule2") == 0) {
        dsvdc_property_add_bytes(property, name, gIconWind16Data, gIconWind16Size);
      } else if (strcmp(dev->mod->type, "NAModule3") == 0) {
        dsvdc_property_add_bytes(property, name, gIconRain16Data, gIconRain16Size);
      } else if (strcmp(dev->mod->type, "NAModule4") == 0) {
        dsvdc_property_add_bytes(property, name, gIconIndoor16Data, gIconIndoor16Size);
      } else {
        dsvdc_property_add_bytes(property, name, gIconStation16Data, gIconStation16Size);
      }

    } else if (strcmp(name, "deviceIcon48") == 0) {
      if (strcmp(dev->mod->type, "NAModule1") == 0) {
        dsvdc_property_add_bytes(property, name, gIconIndoor48Data, gIconIndoor48Size);
      } else if (strcmp(dev->mod->type, "NAModule2") == 0) {
        dsvdc_property_add_bytes(property, name, gIconWind48Data, gIconWind48Size);
      } else if (strcmp(dev->mod->type, "NAModule3") == 0) {
        dsvdc_property_add_bytes(property, name, gIconRain48Data, gIconRain48Size);
      } else if (strcmp(dev->mod->type, "NAModule4") == 0) {
        dsvdc_property_add_bytes(property, name, gIconIndoor48Data, gIconIndoor48Size);
      } else {
        dsvdc_property_add_bytes(property, name, gIconStation48Data, gIconStation48Size);
      }

    } else if (strcmp(name, "deviceIconName") == 0) {
      char info[256];
      if (strcmp(dev->mod->type, "NAModule1") == 0) {
        strcpy(info, "netatmo-outdoor-16.png");
      } else if (strcmp(dev->mod->type, "NAModule2") == 0) {
        strcpy(info, "netatmo-wind-16.png");
      } else if (strcmp(dev->mod->type, "NAModule3") == 0) {
        strcpy(info, "netatmo-rain-16.png");
      } else if (strcmp(dev->mod->type, "NAModule4") == 0) {
        strcpy(info, "netatmo-indoor-16.png");
      } else if (strcmp(dev->mod->type, "NAMain") == 0) {
        strcpy(info, "netatmo-base-16.png"); // main modules
      } else {
        strcpy(info, "netatmo-unknown.png");
      }
      dsvdc_property_add_string(property, name, info);

    } else {
      vdc_report(LOG_WARNING, "get property handler: unhandled name=\"%s\"\n", name);
    }

    free(name);
  }

  pthread_mutex_unlock(&g_network_mutex);
  dsvdc_send_get_property_response(handle, property);
}
